/*
* Copyright (C) 2017      Ugubok
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "AIMP.h"


HRESULT MakeString(IAIMPCore* aimpCore, PWCHAR strSeq, IAIMPString** out)
{
	auto result = aimpCore->CreateObject(IID_IAIMPString, (void**)out);
	if (SUCCEEDED(result))
	{
		return (*out)->SetData(strSeq, wcslen(strSeq));
	}
	return result;
}

IAIMPString* MakeString(IAIMPCore* aimpCore, PWCHAR strSeq) {
	IAIMPString* out;
	MakeString(aimpCore, strSeq, &out);
	return out;
}

/////////////////////////////////////////////////////////////////////
// CAIMPNSFFileFormat
/////////////////////////////////////////////////////////////////////

CAIMPNSFFileFormat::CAIMPNSFFileFormat(IAIMPCore *Core)
{
	aimpCore = Core;
	aimpCore->AddRef();
}

CAIMPNSFFileFormat::~CAIMPNSFFileFormat()
{
	aimpCore->Release();
	aimpCore = nullptr;
}

HRESULT WINAPI CAIMPNSFFileFormat::GetDescription(IAIMPString **S)
{
	MakeString(aimpCore, PLUGIN_FILEFORMAT_DESCRIPTION, S);
	return S_OK;
}

HRESULT WINAPI CAIMPNSFFileFormat::GetExtList(IAIMPString ** S)
{
	MakeString(aimpCore, PLUGIN_FILEFORMAT_EXTLIST, S);
	return S_OK;
}

HRESULT WINAPI CAIMPNSFFileFormat::GetFlags(DWORD ** S)
{
	static DWORD flags = static_cast<DWORD>(AIMP_SERVICE_FILEFORMATS_CATEGORY_AUDIO);
	*S = &flags;
	return S_OK;
}


/////////////////////////////////////////////////////////////////////
// CAimpNSFPlugin
/////////////////////////////////////////////////////////////////////

PWCHAR WINAPI CAIMPNSFPlugin::InfoGet(int Index) 
{
	switch (Index)
	{
	case AIMP_PLUGIN_INFO_NAME:
		return PLUGIN_NAME;

	case AIMP_PLUGIN_INFO_AUTHOR:
		return PLUGIN_AUTHOR;

	case AIMP_PLUGIN_INFO_SHORT_DESCRIPTION:
		return PLUGIN_SHORT_DESC;

	case AIMP_PLUGIN_INFO_FULL_DESCRIPTION:
		return PLUGIN_FULL_DESC;

	default:
		return NULL;
	}
}

DWORD WINAPI CAIMPNSFPlugin::InfoGetCategories() 
{
	return AIMP_PLUGIN_CATEGORY_DECODERS;
}

HRESULT WINAPI CAIMPNSFPlugin::Initialize(IAIMPCore* Core)
{
	aimpCore = Core;
	aimpCore->AddRef();

	if (!IsServiceAvailable(aimpCore, IID_IAIMPServiceFileFormats)
		|| !IsServiceAvailable(aimpCore, IID_IAIMPServiceAudioDecoders))
	{
		aimpCore->Release();
		return E_FAIL;
	}

	decoderExtension = new CAIMPNSFDecoderExtension(aimpCore);
	fileformat = new CAIMPNSFFileFormat(aimpCore);
	bool result;

	decoderExtension->AddRef();
	fileformat->AddRef();

	result = aimpCore->RegisterExtension(IID_IAIMPServiceAudioDecoders, decoderExtension);
	if (!SUCCEEDED(result)) {
#ifdef AIMP_FAIL_MESSAGES
		MessageBoxW(
			NULL,
			L"Failed to register extension IID_IAIMPServiceAudioDecoders", L"NotSo Fatso: FAIL",
			MB_OK | MB_ICONERROR);
#endif
		Finalize();
		return E_ABORT;
	}

	result = aimpCore->RegisterExtension(IID_IAIMPServiceFileFormats, fileformat);
	if (!SUCCEEDED(result)) {
#ifdef AIMP_FAIL_MESSAGES
		MessageBoxW(
			NULL,
			L"Failed to register extension IID_IAIMPServiceFileFormats", L"NotSo Fatso: FAIL",
			MB_OK | MB_ICONERROR);
#endif
		Finalize();
		return E_ABORT;
	}

	return S_OK;
}

HRESULT WINAPI CAIMPNSFPlugin::Finalize()
{
	decoderExtension->Release();
	fileformat->Release();
	aimpCore->Release();
	aimpCore = nullptr;
	delete fileformat;
	delete decoderExtension;
	return S_OK;
}

void CAIMPNSFPlugin::SystemNotification(int NotifyID, IUnknown * Data)
{
}

bool CAIMPNSFPlugin::IsServiceAvailable(IUnknown* provider, REFIID serviceIid)
{
	IUnknown* dummyService;
	if (SUCCEEDED(provider->QueryInterface(serviceIid, (void**)&dummyService)))
	{
		dummyService->Release();
		return true;
	}
	return false;
}


/////////////////////////////////////////////////////////////////////
// CAIMPNSFDecoderExtension
/////////////////////////////////////////////////////////////////////

CAIMPNSFDecoderExtension::CAIMPNSFDecoderExtension(IAIMPCore *Core)
{
	aimpCore = Core;
	aimpCore->AddRef();
}

CAIMPNSFDecoderExtension::~CAIMPNSFDecoderExtension()
{
	aimpCore->Release();
	aimpCore = nullptr;
}

HRESULT CAIMPNSFDecoderExtension::CreateDecoder(IAIMPStream * Stream, DWORD Flags, IAIMPErrorInfo * ErrorInfo, IAIMPAudioDecoder ** Decoder)
{
	int bCreatedSuccessfully;
	CAIMPNSFDecoder* decoder = new CAIMPNSFDecoder(Stream, aimpCore, &bCreatedSuccessfully);

	if (bCreatedSuccessfully)
	{
		*Decoder = decoder;
		return S_OK;
	} 
	else return E_FAIL;  // Cannot initialise NSF Core or load NSF file (not enough memory or wrong file)
	// TODO: Fill ErrorInfo structure, detect what fails: core or file
}


/////////////////////////////////////////////////////////////////////
// CAIMPNSFDecoder
/////////////////////////////////////////////////////////////////////

CAIMPNSFDecoder::CAIMPNSFDecoder(IAIMPStream* inStream, IAIMPCore* Core, int* bSuccess)
{
	aimpCore = Core;
	//aimpCore->AddRef();  // IDK, shall we AddRef in this constructor? Since noone calling destructor...
	*bSuccess = bCoreInitialized = nsfCore.Initialize();
	if (bCoreInitialized)
	{
		fStream = inStream;
		fStream->AddRef();
		if (!nsfFile.AIMPLoadFile(fStream, 1, 1))  // LoadFile returns 0 on success... Strange
			nsfCore.SetPlaybackOptions(48000, 2);
			if (nsfCore.LoadNSF(&nsfFile))
			{
				// no need in fStream anymore, but i not sure that releasing it at that stage is good idea
				fStream->Release();
				nsfCore.SetTrack(1);
				return;
			}

		// Code below executes when fail
		fStream->Release();
		nsfCore.Destroy();
		nsfFile.Destroy();
		*bSuccess = false;
		bCoreInitialized = false;
	}
}

CAIMPNSFDecoder::~CAIMPNSFDecoder()
{
	//aimpCore->Release();
	aimpCore = NULL;
	if (bCoreInitialized)
	{
		nsfCore.Destroy();
		nsfFile.Destroy();
	}
}

BOOL CAIMPNSFDecoder::GetFileInfo(IAIMPFileInfo * FileInfo)
{
	if (!bCoreInitialized) return false;
	FileInfo->SetValueAsObject(AIMP_FILEINFO_PROPID_ARTIST, MakeString(aimpCore, nsfFile.ASArtist));
	FileInfo->SetValueAsObject(AIMP_FILEINFO_PROPID_TITLE, MakeString(aimpCore, nsfFile.ASGameTitle));
	FileInfo->SetValueAsObject(AIMP_FILEINFO_PROPID_COPYRIGHT, MakeString(aimpCore, nsfFile.ASCopyright));
	return true;
}

BOOL CAIMPNSFDecoder::GetStreamInfo(int * SampleRate, int * Channels, int * SampleFormat)
{
	*SampleRate = 48000;
	*Channels = 2;
	*SampleFormat = AIMP_DECODER_SAMPLEFORMAT_16BIT;
	return 0;
}

BOOL CAIMPNSFDecoder::IsSeekable()
{
	return false;
}

BOOL CAIMPNSFDecoder::IsRealTimeStream()
{
	return false;
}

INT64 CAIMPNSFDecoder::GetAvailableData()
{
	return GetSize() - GetPosition();
}

INT64 CAIMPNSFDecoder::GetSize()
{
	//return 0;
	return 122 * 48000 * 2 * (16 / 8);  // 2 min track + 2 sec fade, 48KHz, 2 Channels, 16 bit samples
}

INT64 CAIMPNSFDecoder::GetPosition()
{
	return nsfCore.GetWrittenTime() * 48000 * 2 * (16 / 8);
}

BOOL CAIMPNSFDecoder::SetPosition(const INT64 Value)
{
	//nsfCore.SetWrittenTime((Value / (48000 * 2 * (16 / 8))) * 1000);
	return false;
}

int CAIMPNSFDecoder::Read(void * Buffer, int Count)
{
	if (!bCoreInitialized) return 0;
	return nsfCore.GetSamples(reinterpret_cast<BYTE*>(Buffer), Count);
}


/////////////////////////////////////////////////////////////////////
// CNSFFileModified
/////////////////////////////////////////////////////////////////////
// All code in this section comes from NSF_File.cpp, and modified for AIMP buffer interface support

// Defines from NSF_File.cpp
#define SAFE_DELETE(p) { if(p){ delete[] p; p = NULL; } }
#define SAFE_NEW(p,t,s,r) p = new t[s]; if(!p) return r; ZeroMemory(p,sizeof(t) * s)

// Macro for AIMP buffer interface, that works like feof()
#define FEOF(s) (s->GetPosition() >= s->GetSize())
// Macro for casting pointers to type that AIMP buffer interface expects
#define _PUCHAR(p) (reinterpret_cast<unsigned char*>(p))

void CNSFFileModified::Destroy()
{
	CNSFFile::Destroy();
	if (ASGameTitle) {
		SAFE_DELETE(ASGameTitle); ASGameTitle = NULL;
	}
	if (ASArtist) {
		SAFE_DELETE(ASArtist); ASArtist = NULL;
	}
	if (ASCopyright) {
		SAFE_DELETE(ASCopyright); ASCopyright = NULL;
	}
	if (ASRipper) {
		SAFE_DELETE(ASRipper); ASRipper = NULL;
	}
}

int CNSFFileModified::AIMPLoadFile(IAIMPStream * fStream, BYTE needdata, BYTE ignoreversion)
{
	Destroy();
	fStream->Seek(0, AIMP_STREAM_SEEKMODE_FROM_BEGINNING);

	UINT type = 0;
	fStream->Read(_PUCHAR(&type), 4);
	int ret = -1;

	if (type == HEADERTYPE_NESM)		ret = AIMPLoadFile_NESM(fStream, needdata, ignoreversion);
	if (type == HEADERTYPE_NSFE)		ret = AIMPLoadFile_NSFE(fStream, needdata);

	// Snake's revenge puts '00' for the initial track, which (after subtracting 1) makes it 256 or -1 (bad!)
	// This prevents that crap
	if (nInitialTrack >= nTrackCount)
		nInitialTrack = 0;
	if (nInitialTrack < 0)
		nInitialTrack = 0;

	// if there's no tracks... this is a crap NSF
	if (nTrackCount < 1)
	{
		Destroy();
		return -1;
	}

	return ret;
}

int CNSFFileModified::AIMPLoadFile_NESM(IAIMPStream * fStream, BYTE needdata, BYTE ignoreversion)
{
	INT64 len;

	len = fStream->GetSize() - 0x80;
	fStream->Seek(0, AIMP_STREAM_SEEKMODE_FROM_BEGINNING);

	if (len < 1) return -1;

	//read the info
	NESM_HEADER	hdr;
	fStream->Read(_PUCHAR(&hdr), 0x80);

	//confirm the header
	if (hdr.nHeader != HEADERTYPE_NESM)			return -1;
	if (hdr.nHeaderExtra != 0x1A)				return -1;
	if ((!ignoreversion) && (hdr.nVersion != 1))	return -1; //stupid NSFs claim to be above version 1  >_>
	//NESM is generally easier to work with (but limited!)
	//  just move the data over from NESM_HEADER over to our member data

	bIsExtended = 0;
	nIsPal = hdr.nNTSC_PAL & 0x03;
	nPAL_PlaySpeed = hdr.nSpeedPAL;			//blarg
	nNTSC_PlaySpeed = hdr.nSpeedNTSC;		//blarg
	nLoadAddress = hdr.nLoadAddress;
	nInitAddress = hdr.nInitAddress;
	nPlayAddress = hdr.nPlayAddress;
	nChipExtensions = hdr.nExtraChip;


	nTrackCount = hdr.nTrackCount;
	nInitialTrack = hdr.nInitialTrack - 1;	//stupid 1-based number =P

	memcpy(nBankswitch, hdr.nBankSwitch, 8);

	SAFE_NEW(szGameTitle, char, 33, 1);
	SAFE_NEW(szArtist, char, 33, 1);
	SAFE_NEW(szCopyright, char, 33, 1);
	SAFE_NEW(ASGameTitle, WCHAR, 33, 1);
	SAFE_NEW(ASArtist, WCHAR, 33, 1);
	SAFE_NEW(ASCopyright, WCHAR, 33, 1);

	memcpy(szGameTitle, hdr.szGameTitle, 32);
	memcpy(szArtist, hdr.szArtist, 32);
	memcpy(szCopyright, hdr.szCopyright, 32);
	MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, szGameTitle, -1, ASGameTitle, 33);
	MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, szArtist,	 -1, ASArtist, 33);
	MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, szCopyright, -1, ASCopyright, 33);

	//read the NSF data
	if (needdata)
	{
		SAFE_NEW(pDataBuffer, BYTE, static_cast<unsigned int>(len), 1);
		fStream->Read(pDataBuffer, static_cast<unsigned int>(len));
		nDataBufferSize = static_cast<unsigned int>(len);
	}

	//if we got this far... it was a successful read
	return 0;
}

int CNSFFileModified::AIMPLoadFile_NSFE(IAIMPStream * fStream, BYTE needdata)
{
	//restart the file
	fStream->Seek(0, AIMP_STREAM_SEEKMODE_FROM_BEGINNING);

	//the vars we'll be using
	UINT nChunkType;
	int  nChunkSize;
	int  nChunkUsed;
	INT64	nDataPos = 0;
	BYTE	bInfoFound = 0;
	BYTE	bEndFound = 0;
	BYTE	bBankFound = 0;

	NSFE_INFOCHUNK	info;
	ZeroMemory(&info, sizeof(NSFE_INFOCHUNK));
	info.nTrackCount = 1;		//default values

	//confirm the header!
	fStream->Read(_PUCHAR(&nChunkType), 4);
	if (nChunkType != HEADERTYPE_NSFE)			return -1;

	//begin reading chunks
	while (!bEndFound)
	{
		if (FEOF(fStream))						return -1;
		fStream->Read(_PUCHAR(&nChunkSize), 4);
		fStream->Read(_PUCHAR(&nChunkType), 4);

		switch (nChunkType)
		{
		case CHUNKTYPE_INFO:
			if (bInfoFound)						return -1;	//only one info chunk permitted
			if (nChunkSize < 8)					return -1;	//minimum size

			bInfoFound = 1;
			nChunkUsed = min((int)sizeof(NSFE_INFOCHUNK), nChunkSize);

			fStream->Read(_PUCHAR(&info), nChunkUsed);
			fStream->Seek(nChunkSize - nChunkUsed, AIMP_STREAM_SEEKMODE_FROM_CURRENT);

			bIsExtended = 1;
			nIsPal = info.nIsPal & 3;
			nLoadAddress = info.nLoadAddress;
			nInitAddress = info.nInitAddress;
			nPlayAddress = info.nPlayAddress;
			nChipExtensions = info.nExt;
			nTrackCount = info.nTrackCount;
			nInitialTrack = info.nStartingTrack;

			nPAL_PlaySpeed = (WORD)(1000000 / PAL_NMIRATE);		//blarg
			nNTSC_PlaySpeed = (WORD)(1000000 / NTSC_NMIRATE);	//blarg
			break;

		case CHUNKTYPE_DATA:
			if (!bInfoFound)					return -1;
			if (nDataPos)						return -1;
			if (nChunkSize < 1)					return -1;

			nDataBufferSize = nChunkSize;
			nDataPos = fStream->GetPosition();

			fStream->Seek(nChunkSize, AIMP_STREAM_SEEKMODE_FROM_CURRENT);
			break;

		case CHUNKTYPE_NEND:
			bEndFound = 1;
			break;

		case CHUNKTYPE_TIME:
			if (!bInfoFound)					return -1;
			if (pTrackTime)						return -1;

			SAFE_NEW(pTrackTime, int, nTrackCount, 1);
			nChunkUsed = min(nChunkSize / 4, nTrackCount);

			fStream->Read(_PUCHAR(pTrackTime), nChunkUsed*4);
			fStream->Seek(nChunkSize - (nChunkUsed * 4), AIMP_STREAM_SEEKMODE_FROM_CURRENT);

			for (; nChunkUsed < nTrackCount; nChunkUsed++)
				pTrackTime[nChunkUsed] = -1;	//negative signals to use default time

			break;

		case CHUNKTYPE_FADE:
			if (!bInfoFound)					return -1;
			if (pTrackFade)						return -1;

			SAFE_NEW(pTrackFade, int, nTrackCount, 1);
			nChunkUsed = min(nChunkSize / 4, nTrackCount);

			fStream->Read(_PUCHAR(pTrackFade), nChunkUsed * 4);
			fStream->Seek(nChunkSize - (nChunkUsed * 4), AIMP_STREAM_SEEKMODE_FROM_CURRENT);

			for (; nChunkUsed < nTrackCount; nChunkUsed++)
				pTrackFade[nChunkUsed] = -1;	//negative signals to use default time

			break;

		case CHUNKTYPE_BANK:
			if (bBankFound)						return -1;

			bBankFound = 1;
			nChunkUsed = min(8, nChunkSize);

			fStream->Read(nBankswitch, nChunkUsed);
			fStream->Seek(nChunkSize - nChunkUsed, AIMP_STREAM_SEEKMODE_FROM_CURRENT);
			break;

		case CHUNKTYPE_PLST:
			if (pPlaylist)						return -1;

			nPlaylistSize = nChunkSize;
			if (nPlaylistSize < 1)				break;  //no playlist?

			SAFE_NEW(pPlaylist, BYTE, nPlaylistSize, 1);
			fStream->Read(pPlaylist, nChunkSize);
			break;

		case CHUNKTYPE_AUTH: {
			if (szGameTitle)						return -1;

			char*		buffer;
			char*		ptr;
			SAFE_NEW(buffer, char, nChunkSize + 4, 1);

			fStream->Read(_PUCHAR(buffer), nChunkSize);
			ptr = buffer;

			char**		ar[4] = { &szGameTitle, &szArtist, &szCopyright, &szRipper };
			WCHAR**		wr[4] = { &ASGameTitle, &ASArtist, &ASCopyright, &ASRipper };

			int			i;
			for (i = 0; i < 4; i++)
			{
				nChunkUsed = lstrlen(ptr) + 1;
				*ar[i] = new char[nChunkUsed];
				*wr[i] = new WCHAR[nChunkUsed];
				if (!*ar[i] || !*wr[i]) { SAFE_DELETE(buffer); return 0; }
				memcpy(*ar[i], ptr, nChunkUsed);
				MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, ptr, -1, *wr[i], nChunkUsed);
				ptr += nChunkUsed;
			}
			SAFE_DELETE(buffer);
		}break;

		case CHUNKTYPE_TLBL: {
			if (!bInfoFound)						return -1;
			if (szTrackLabels)						return -1;

			SAFE_NEW(szTrackLabels, char*, nTrackCount, 1);

			char*		buffer;
			char*		ptr;
			SAFE_NEW(buffer, char, nChunkSize + nTrackCount, 1);

			fStream->Read(_PUCHAR(buffer), nChunkSize);
			ptr = buffer;

			int			i;
			for (i = 0; i < nTrackCount; i++)
			{
				nChunkUsed = lstrlen(ptr) + 1;
				szTrackLabels[i] = new char[nChunkUsed];
				if (!szTrackLabels[i]) { SAFE_DELETE(buffer); return 0; }
				memcpy(szTrackLabels[i], ptr, nChunkUsed);
				ptr += nChunkUsed;
			}
			SAFE_DELETE(buffer);
		}break;

		default:		//unknown chunk
			nChunkType &= 0x000000FF;  //check the first byte
			if ((nChunkType >= 'A') && (nChunkType <= 'Z'))	//chunk is vital... don't continue
				return -1;
			//otherwise, just skip it
			fStream->Seek(nChunkSize, AIMP_STREAM_SEEKMODE_FROM_CURRENT);

			break;
		}		//end switch
	}			//end while

				//if we exited the while loop without a 'return', we must have hit an NEND chunk
				//  if this is the case, the file was layed out as it was expected.
				//  now.. make sure we found both an info chunk, AND a data chunk... since these are
				//  minimum requirements for a valid NSFE file

	if (!bInfoFound)			return -1;
	if (!nDataPos)				return -1;

	//if both those chunks existed, this file is valid.  Load the data if it's needed

	if (needdata)
	{
		fStream->Seek(nDataPos, AIMP_STREAM_SEEKMODE_FROM_BEGINNING);
		SAFE_NEW(pDataBuffer, BYTE, nDataBufferSize, 1);
		fStream->Read(pDataBuffer, nDataBufferSize);
	}
	else
		nDataBufferSize = 0;

	//return success!
	return 0;
}

#undef SAFE_NEW
#undef SAFE_DELETE
#undef FEOF
#undef _PUCHAR

/////////////////////////////////////////////////////////////////////
// EXPORT
/////////////////////////////////////////////////////////////////////

bool WINAPI AIMPPluginGetHeader(IAIMPPlugin ** Header)
{
	auto plugin = new CAIMPNSFPlugin();
	plugin->AddRef();
	*Header = plugin;
	return true;
}